@@ -8,7 +8,7 @@ use std::{
 use tanoshi_lib::prelude::{Chapter, Extension, ExtensionResult, Filters, Manga, Param, Source};
 use tanoshi_util::http::Request;
 use tokio::{
-    sync::mpsc::{UnboundedReceiver, UnboundedSender},
+    sync::mpsc::{Receiver, Sender},
     task::JoinHandle,
     time::Instant,
 };
@@ -26,16 +26,19 @@ struct ExtensionEnv {
 }
 
 pub struct ExtensionProxy {
-    instance: Instance,
-    env: ExtensionEnv,
+    path: PathBuf,
 }
 
 impl ExtensionProxy {
-    pub fn load<P: AsRef<Path>>(
-        store: &Store,
-        path: P,
-    ) -> Result<Arc<dyn Extension>, Box<dyn std::error::Error>> {
-        let module = unsafe { Module::deserialize_from_file(store, path)? };
+    pub fn new<P: AsRef<Path>>(path: P) -> Arc<dyn Extension> {
+        let path = PathBuf::new().join(path);
+        Arc::new(Self { path })
+    }
+
+    fn load(&self) -> Result<(Instance, ExtensionEnv), Box<dyn std::error::Error>> {
+        let instant = Instant::now();
+        let store = Self::init_store_headless();
+        let module = unsafe { Module::deserialize_from_file(&store, &self.path)? };
 
         let stdin = Pipe::new();
         let stdout = Pipe::new();
@@ -53,18 +56,18 @@ impl ExtensionProxy {
 
         let tanoshi = imports! {
             "tanoshi" => {
-                "host_http_request" => Function::new_native_with_env(store, env.clone(), host_http_request),
-                "host_debug" => Function::new_native_with_env(store, env.clone(), host_debug),
-                "host_error" => Function::new_native_with_env(store, env.clone(), host_error),
-                "host_info" => Function::new_native_with_env(store, env.clone(), host_info),
-                "host_trace" => Function::new_native_with_env(store, env.clone(), host_trace),
-                "host_warn" => Function::new_native_with_env(store, env.clone(), host_warn),
+                "host_http_request" => Function::new_native_with_env(&store, env.clone(), host_http_request),
+                "host_debug" => Function::new_native_with_env(&store, env.clone(), host_debug),
+                "host_error" => Function::new_native_with_env(&store, env.clone(), host_error),
+                "host_info" => Function::new_native_with_env(&store, env.clone(), host_info),
+                "host_trace" => Function::new_native_with_env(&store, env.clone(), host_trace),
+                "host_warn" => Function::new_native_with_env(&store, env.clone(), host_warn),
             }
         };
 
         let instance = Instance::new(&module, &tanoshi.chain_back(import_object))?;
-
-        Ok(Arc::new(ExtensionProxy { instance, env }))
+        debug!("extension loaded in {} ms", instant.elapsed().as_millis());
+        Ok((instance, env))
     }
 
     #[cfg(feature = "compiler")]
@@ -118,9 +121,10 @@ impl ExtensionProxy {
     where
         T: DeserializeOwned,
     {
-        let res = self.instance.exports.get_function(name)?;
+        let (instance, env) = self.load()?;
+        let res = instance.exports.get_function(name)?;
         res.call(&[])?;
-        let object_str = wasi_read(&self.env)?;
+        let object_str = wasi_read(&env)?;
         debug!("call {} => {}", name, object_str);
         Ok(ron::from_str(&object_str)?)
     }
@@ -130,12 +134,13 @@ impl ExtensionProxy {
         T: DeserializeOwned,
         U: Serialize + Debug,
     {
-        let res = self.instance.exports.get_function(name)?;
-        if let Err(e) = wasi_write(&self.env, &param) {
+        let (instance, env) = self.load()?;
+        let res = instance.exports.get_function(name)?;
+        if let Err(e) = wasi_write(&env, &param) {
             error!("error write to wasi: {}", e);
         }
         res.call(&[])?;
-        let object_str = wasi_read(&self.env)?;
+        let object_str = wasi_read(&env)?;
         debug!("call {}({:?}) => {}", name, param, object_str);
         Ok(ron::from_str(&object_str)?)
     }
@@ -182,11 +187,10 @@ impl Extension for ExtensionProxy {
     }
 }
 
-pub fn start<P: AsRef<Path>>(path: P) -> (JoinHandle<()>, UnboundedSender<Command>) {
-    let (tx, rx) = tokio::sync::mpsc::unbounded_channel();
-    let path = PathBuf::new().join(path);
+pub fn start() -> (JoinHandle<()>, Sender<Command>) {
+    let (tx, rx) = tokio::sync::mpsc::channel(25);
     let handle = tokio::spawn(async move {
-        thread_main(path, rx).await;
+        thread_main(rx).await;
     });
 
     (handle, tx)
@@ -194,9 +198,9 @@ pub fn start<P: AsRef<Path>>(path: P) -> (JoinHandle<()>, UnboundedSender<Comman
 
 pub async fn load<P: AsRef<Path>>(
     path: P,
-    tx: UnboundedSender<Command>,
+    tx: Sender<Command>,
 ) -> Result<(), Box<dyn std::error::Error>> {
-    match std::fs::read_dir(&path) {
+    match tokio::fs::read_dir(&path).await {
         Ok(_) => {}
         Err(_) => {
             let _ = std::fs::create_dir_all(&path);
@@ -206,53 +210,22 @@ pub async fn load<P: AsRef<Path>>(
     #[cfg(feature = "compiler")]
     compile(&path).await?;
 
-    for entry in std::fs::read_dir(&path)?
-        .into_iter()
-        .filter_map(Result::ok)
-        .filter(move |path| {
-            path.path()
-                .extension()
-                .map_or(false, |ext| ext == "tanoshi")
-        })
-    {
+    let mut read_dir = tokio::fs::read_dir(&path).await?;
+    while let Some(entry) = read_dir.next_entry().await? {
+        if !entry
+            .path()
+            .extension()
+            .map_or(false, |ext| ext == "tanoshi")
+        {
+            continue;
+        }
+
         let path = entry.path();
         info!("found compiled plugin at {:?}", path.clone());
         tx.send(Command::Load(
             path.to_str().ok_or("no path str")?.to_string(),
-        ))?;
-    }
-
-    Ok(())
-}
-
-pub async fn reload<P: AsRef<Path>>(
-    path: P,
-    store: &Store,
-    extension_map: &mut BTreeMap<i64, (Source, Arc<dyn Extension>)>,
-) -> Result<(), Box<dyn std::error::Error>> {
-    let path = PathBuf::new().join(path);
-    for entry in std::fs::read_dir(&path)?
-        .into_iter()
-        .filter_map(Result::ok)
-        .filter(move |path| {
-            path.path()
-                .extension()
-                .map_or(false, |ext| ext == "tanoshi")
-        })
-    {
-        let path = entry.path();
-        info!("load plugin from {:?}", path.display());
-        let now = Instant::now();
-        match ExtensionProxy::load(store, &path) {
-            Ok(proxy) => {
-                let source = proxy.detail();
-                info!("loaded in {} ms: {:?}", now.elapsed().as_millis(), source);
-                extension_map.insert(source.id, (source, proxy));
-            }
-            Err(e) => {
-                error!("error load extension: {}", e);
-            }
-        }
+        ))
+        .await?;
     }
 
     Ok(())
@@ -280,37 +253,40 @@ pub async fn compile_with_target<P: AsRef<Path>>(
     let cpu_feature = CpuFeature::set();
     let target = Target::new(triple, cpu_feature);
 
-    match std::fs::read_dir(&path) {
+    match tokio::fs::read_dir(&path).await {
         Ok(_) => {}
         Err(_) => {
             let _ = std::fs::create_dir_all(&path);
         }
     }
 
-    for entry in std::fs::read_dir(&path)?
-        .into_iter()
-        .filter_map(Result::ok)
-        .filter(move |path| path.path().extension().map_or(false, |ext| ext == "wasm"))
-    {
+    let mut read_dir = tokio::fs::read_dir(&path).await?;
+    while let Some(entry) = read_dir.next_entry().await? {
+        if !entry
+            .path()
+            .extension()
+            .map_or(false, |ext| ext == "tanoshi")
+        {
+            continue;
+        }
+
         let path = entry.path();
         info!("found wasm file at {:?}", path.clone());
         ExtensionProxy::compile_from_file(&path, target.clone()).map_err(|e| format!("{}", e))?;
 
         if remove_wasm {
             debug!("remove wasm file");
-            std::fs::remove_file(path)?;
+            tokio::fs::remove_file(path).await?;
         }
     }
 
     Ok(())
 }
 
-async fn thread_main<P: AsRef<Path>>(path: P, extension_receiver: UnboundedReceiver<Command>) {
+async fn thread_main(extension_receiver: Receiver<Command>) {
     let mut recv = extension_receiver;
     let mut extension_map: BTreeMap<i64, (Source, Arc<dyn Extension>)> = BTreeMap::new();
 
-    let mut store = ExtensionProxy::init_store_headless();
-
     while let Some(cmd) = recv.recv().await {
         match cmd {
             Command::Insert(source_id, proxy) => {
@@ -320,46 +296,13 @@ async fn thread_main<P: AsRef<Path>>(path: P, extension_receiver: UnboundedRecei
             Command::Load(path) => {
                 info!("load plugin from {:?}", path);
                 let now = Instant::now();
-                match ExtensionProxy::load(&store, &path) {
-                    Ok(proxy) => {
-                        let source = proxy.detail();
-                        info!("loaded in {} ms: {:?}", now.elapsed().as_millis(), source);
-                        extension_map.insert(source.id, (source, proxy));
-                    }
-                    Err(e) => {
-                        error!("error load extension: {}", e);
-                    }
-                }
+                let proxy = ExtensionProxy::new(&path);
+                let source = proxy.detail();
+                info!("loaded in {} ms: {:?}", now.elapsed().as_millis(), source);
+                extension_map.insert(source.id, (source, proxy));
             }
             Command::Unload(source_id, tx) => {
-                let extension_path = if let Some((source, _)) = extension_map.get_mut(&source_id) {
-                    PathBuf::new()
-                        .join(&path)
-                        .join(&source.name)
-                        .with_extension("tanoshi")
-                } else {
-                    continue;
-                };
-
-                // drop extension map
-                drop(extension_map);
-                // drop extension map
-                drop(store);
-
-                // remove the file
-                if let Err(e) = std::fs::remove_file(&extension_path) {
-                    error!("error removing {}: {}", extension_path.display(), e);
-                }
-
-                // assign new extension map
-                extension_map = BTreeMap::new();
-                // initialize new store
-                store = ExtensionProxy::init_store_headless();
-
-                // reload all extension
-                if let Err(e) = reload(&path, &store, &mut extension_map).await {
-                    error!("error reloading extension: {}", e);
-                }
+                drop(extension_map.remove(&source_id));
 
                 if tx.send(()).is_err() {
                     error!("[Command::Unload] receiver dropped");
@@ -393,31 +336,34 @@ async fn thread_main<P: AsRef<Path>>(path: P, extension_receiver: UnboundedRecei
                 }
             },
             Command::Filters(source_id, tx) => {
-                process(&extension_map, source_id, tx, |proxy| proxy.filters());
+                process(&extension_map, source_id, tx, |proxy| proxy.filters()).await;
             }
             Command::GetMangaList(source_id, param, tx) => {
                 process(&extension_map, source_id, tx, |proxy| {
                     proxy.get_manga_list(param)
-                });
+                })
+                .await;
             }
             Command::GetMangaInfo(source_id, path, tx) => {
                 process(&extension_map, source_id, tx, |proxy| {
                     proxy.get_manga_info(path)
-                });
+                })
+                .await;
             }
             Command::GetChapters(source_id, path, tx) => {
                 process(&extension_map, source_id, tx, |proxy| {
                     proxy.get_chapters(path)
-                });
+                })
+                .await;
             }
             Command::GetPages(source_id, path, tx) => {
-                process(&extension_map, source_id, tx, |proxy| proxy.get_pages(path));
+                process(&extension_map, source_id, tx, |proxy| proxy.get_pages(path)).await;
             }
         }
     }
 }
 
-fn process<F, T>(
+async fn process<F, T>(
     extension_map: &BTreeMap<i64, (Source, Arc<dyn Extension>)>,
     source_id: i64,
     tx: ExtensionResultSender<T>,
@@ -429,10 +375,14 @@ fn process<F, T>(
     match extension_map.get(&source_id) {
         Some(proxy) => {
             let (_, proxy) = proxy.clone();
-            let res = f(proxy);
-            if tx.send(res).is_err() {
-                error!("[process] receiver dropped");
-            }
+            tokio::task::spawn_blocking(move || {
+                let res = f(proxy);
+                if tx.send(res).is_err() {
+                    error!("[process] receiver dropped");
+                }
+            })
+            .await
+            .expect("failed to spawn process");
         }
         None => {
             error!("extension with id {} not found", source_id);
